#!/usr/bin/env python3.7
import os
import sys
import yaml
import uvloop
import asyncio
from nteu_translation_engine.engine import Engine
from aiohttp import web
from typing import Dict


asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


class TranslationEngineServer:

    def __init__(self, config: str):
        with open(config, "r") as file:
            config = yaml.safe_load(file)

        self._update_config(config)

        self._app = web.Application()

        self._app["lock"] = asyncio.Lock()
        self._app["engine"] = Engine(config)

        self._app.router.add_post("/translate", handle_translate)

    def start(self, port: int):
        web.run_app(self._app, port=port)

    def _update_config(self, root: Dict, prefix: str = None):
        for key in root.keys():
            name = prefix + key.upper() if prefix is not None else key.upper()
            print(name)
            value = root[key]
            if isinstance(value, Dict):
                self._update_config(value, name + "_")
            elif os.environ.get(name) is not None:
                root[key] = os.environ.get(name)


async def handle_translate(req):
    lock = req.app["lock"]
    engine = req.app["engine"]

    try:
        req = await req.json()
        batch = req.get("srcs")
        translations = await engine.process_batch(batch, lock)
        ans = {
            "tus": []
        }
        for src, translation in zip(batch, translations):
            ans["tus"].append({"src": src, "tgt": translation})
        return web.json_response(ans, status=200)

    except Exception as e:
        response_obj = {'status': 'failed', 'reason': str(e)}
        return web.json_response(response_obj, status=500)


if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) < 2:
        print("Three arguments needed: ")
        print("config and port")
    else:
        try:
            server = TranslationEngineServer(args[0])
            server.start(int(args[1]))
        except Exception as e:
            print("Something went wrong: ", str(e))
            print("Three arguments needed: ")
            print("config and port")



from pangeamt_nlp.processor.pipeline_decoding import PipelineDecoding
from pangeamt_nlp.truecaser.truecaser import Truecaser
from pangeamt_nlp.bpe.bpe import BPE
from pangeamt_nlp.tokenizer.tokenizer_factory import TokenizerFactory
from pangeamt_nlp.seg import Seg
from typing import Dict
import os
import logging
from logging import Logger


class Pipeline:

    def __init__(self, config: Dict, logger: Logger = None) -> "Pipeline":
        self._config = config
        self._src_lang = config["src_lang"]
        self._tgt_lang = config["tgt_lang"]
        self._processors = self.load_decoding_pipeline()
        self._src_bpe = self.load_bpe()
        self._src_truecaser, self._tgt_truecaser = self.load_truecaser()
        self._src_tokenizer, self._tgt_tokenizer = self.load_tokenizer()
        self._logger = logger

    def log(self, msg: str, lvl: int = logging.INFO):
        if self._logger is not None:
            self._logger.log(lvl, msg)

    async def preprocess(self, seg: Seg):
        self._processors.process_src(seg, logger=self._logger)
        self.log("Tokenizing", logging.DEBUG)
        seg.src = self._src_tokenizer.tokenize(seg.src)
        self.log(seg.src, logging.DEBUG)
        if self._src_truecaser is not None:
            self.log("Truecasing", logging.DEBUG)
            seg.src = self._src_truecaser.truecase(seg.src)
            self.log(seg.src, logging.DEBUG)
        if self._src_bpe is not None:
            self.log("BPE", logging.DEBUG)
            seg.src = self._src_bpe.apply(seg.src)
            self.log(seg.src, logging.DEBUG)

    async def postprocess(self, seg: Seg):
        self.log("Undo BPE", logging.DEBUG)
        seg.tgt = BPE.undo(seg.tgt)
        self.log(seg.tgt, logging.DEBUG)
        self.log("Detokenize", logging.DEBUG)
        seg.tgt = self._tgt_tokenizer.detokenize(seg.tgt.split(" "))
        self.log(seg.tgt, logging.DEBUG)
        if self._tgt_truecaser:
            self.log("Detruecasing")
            seg.tgt = self._tgt_truecaser.detruecase(seg.tgt)
            self.log(seg.tgt, logging.DEBUG)
        self._processors.process_tgt(seg, logger=self._logger)

    def load_decoding_pipeline(self):
        return PipelineDecoding.create_from_dict(
            self._src_lang, self._tgt_lang, self._config["processors"]
        )

    def load_bpe(self):
        if self._config["bpe"] is not None:
            path = self._config["translation_engine_server"]["bpe"]
            if self._config["bpe"]["joint"]:
                codes = os.path.join(path, "codes32k.txt")
                vocab = os.path.join(path, "src_vocab.txt")
                return BPE(codes, vocab)
            else:
                codes = os.paht.join(path, "src_codes.txt")
                return BPE(codes)
        return None

    def load_truecaser(self):
        if self._config["truecaser"]["src"] == "enabled":
            path = os.path.join(
                self._config["translation_engine_server"]["truecaser"],
                "src_model.txt"
            )
            truecaser = Truecaser(path)
            return truecaser, truecaser
        elif self._config["truecaser"]["tgt"] == "enabled":
            return None, Truecaser()
        else:
            return None, None

    def load_tokenizer(self):
        src_tok_name = self._config["tokenizer"]["src"]
        tgt_tok_name = self._config["tokenizer"]["tgt"]

        src_tokenizer = TokenizerFactory.new(self._src_lang, src_tok_name)
        tgt_tokenizer = TokenizerFactory.new(self._tgt_lang, tgt_tok_name)

        return (src_tokenizer, tgt_tokenizer)



from setuptools import setup

setup(
    name='nteu_translation_engine',
    version='0.4.6',
    description='NTEU translation engine',
    url='https://github.com/Pangeamt/nteu-translation-engine',
    author='PangeaMT',
    author_email='a.cerda@pangeanic.es',
    license='MIT',
    packages=[
        'nteu_translation_engine'
    ],
    install_requires=[
        "aiohttp==3.6.2",
        "click",
        "pangeamt-nlp",
        "PyYAML",
        "uvloop"
    ],
    zip_safe=False
)



from pangeamt_nlp.translation_model.translation_model_factory import (
    TranslationModelFactory,
)
import logging
from pangeamt_nlp.seg import Seg
from nteu_translation_engine.pipeline import Pipeline
from typing import Dict, List


class Engine:
    def __init__(self, config: Dict, log_file: str = None, dbug: bool = False):
        self._logging = log_file is not None
        if self._logging:
            self._lvl = logging.DEBUG if dbug else logging.INFO
            logging.basicConfig(
                handlers=[logging.FileHandler(log_file)],
                level=self._lvl,
                format="%(asctime)s :: %(levelname)s :: %(message)s"
            )
            self._logger = logging.getLogger("my_logger")
        else:
            self._logger = None
        self._config = config
        self._model = self.load_model()
        self._pipeline = Pipeline(config, self._logger)

    def log(self, message: str, level: str = logging.INFO):
        if self._logging:
            self._logger.log(level, message)

    def load_model(self):
        name = self._config["translation_model"]["name"]
        args = self._config["translation_model"]["args_decoding"]
        if self._config["translation_engine_server"]["gpu"]:
            args["gpu"] = 0
        model_path = self._config["translation_engine_server"]["model_path"]
        self.log(f"Loading -> {name} with arguments {args}.")
        translation_model = TranslationModelFactory.get_class(name)
        return translation_model(model_path, **args)

    async def translate(self, srcs: List):
        translations = self._model.translate(srcs)
        result = []
        for translation in translations:
            if self._lvl == logging.DEBUG:
                n_best = len(translation)
                log_msg = (
                    f"For sentence {translation.pred_sents[0].src_raw}",
                    f"{n_best}-best translations:"
                )
                self.log(log_msg, level=logging.DEBUG)
                f = "\nOption #{0} with score {1}:\n{2}\nAttention:\n"
                for i, prediction in enumerate(translation.pred_sents):
                    log_msg = f.format(
                        i,
                        prediction.score,
                        prediction.sentence
                    )
                    table = prediction.get_pretty_attention()
                    if table is not None:
                        table_fields = table._field_names
                        table_width = len(table_fields)
                        for column_index in range(0, table_width, 3):
                            f_index = column_index
                            l_index = min([column_index + 3, table_width])
                            fields_to_take = table_fields[f_index:l_index]
                            str_to_add = table.get_string(
                                fields=fields_to_take
                            ) + "\n"
                            log_msg += str_to_add
                    self.log(log_msg, level=logging.DEBUG)
            result.append(translation.pred_sents[0].sentence)
        return result

    async def process_batch(self, batch: List, lock=None):
        srcs = []
        segs = []
        ans = []
        for src in batch:
            seg = Seg(src)
            await self._pipeline.preprocess(seg)
            srcs.append(seg.src)
            segs.append(seg)
        if lock is not None:
            async with lock:
                translations = await self.translate(srcs)
        else:
            translations = await self.translate(srcs)
        for translation, seg in zip(translations, segs):
            seg.tgt = seg.tgt_raw = translation
            await self._pipeline.postprocess(seg)
            ans.append(seg.tgt)
            self.log(
                f"Translated -> {seg.src_raw} -> {seg.src} "
                f"-> {seg.tgt_raw} -> {seg.tgt}"
            )
        return ans
